<?php
/**
* @version $Id: cbpaidsubscriptions.quickpay.php 1581 2012-12-24 02:36:44Z beat $
* @package CBSubs (TM) Community Builder Plugin for Paid Subscriptions (TM)
* @subpackage Plugin for Paid Subscriptions
* @copyright (C) 2007-2019 and Trademark of Lightning MultiCom SA, Switzerland - www.joomlapolis.com - and its licensors, all rights reserved
* @license http://www.gnu.org/licenses/old-licenses/gpl-2.0.html GNU/GPL version 2
*/

use CBLib\Registry\ParamsInterface;

/** ensure this file is being included by a parent file */
if ( ! ( defined( '_VALID_CB' ) || defined( '_JEXEC' ) || defined( '_VALID_MOS' ) ) ) { die( 'Direct Access to this location is not allowed.' ); }

global $_CB_framework;

// Avoids errors in CB plugin edit:
/** @noinspection PhpIncludeInspection */
include_once( $_CB_framework->getCfg( 'absolute_path' ) . '/components/com_comprofiler/plugin/user/plug_cbpaidsubscriptions/cbpaidsubscriptions.class.php');

// This gateway implements a payment handler using a hosted page at the PSP:
// Import class cbpaidHostedPagePayHandler that extends cbpaidPayHandler
// and implements all gateway-generic CBSubs methods.

/**
 * Payment handler class for this gateway: Handles all payment events and notifications, called by the parent class:
 *
 * This is the OEM version
 * Please note that except the constructor and the API version this class does not implement any public methods.
 */
class cbpaidquickpayoem extends cbpaidHostedPagePayHandler
{
	/**
	 * Gateway API version used - Not used anywhere !
	 * @var int
	 */
	public $gatewayApiVersion	=	"v10";

	/**
	 * Constructor
	 *
	 * @param cbpaidGatewayAccount $account
	 */
	public function __construct( $account )
	{
		parent::__construct( $account );

		// Set gateway URLS for $this->pspUrl() results: first 2 are the main hosted payment page posting URL, next ones are gateway-specific:
		$this->_gatewayUrls	=	array(	'psp+normal'		=>	$this->getAccountParam( 'psp_normal_url' ),
		                                'psp+test'			=>	$this->getAccountParam( 'psp_test_url' ),
		                                'psp+api+normal'	=>	$this->getAccountParam( 'psp_api_normal_url' ),
		                                'psp+api+test'		=>	$this->getAccountParam( 'psp_api_test_url' ) );
	}

	/**
	 * CBSUBS HOSTED PAGE PAYMENT API METHODS:
	 */

	/**
	 * Returns single payment request parameters for gateway depending on basket (without specifying payment type)
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket   paymentBasket object
	 * @return array                                 Returns array $requestParams
	 */
	protected function getSinglePaymentRequstParams( $paymentBasket )
	{
		// build hidden form fields or redirect to gateway url parameters array:
		$requestParams						=	$this->_getBasicRequstParams( $paymentBasket );

		// Add 3D Secure option
		if($this->getAccountParam( '3d_secure' ) == '1'){
			$requestParams['payment_methods']       = '3d-creditcard,mobilepay';
		}
		// sign single payment params:
		$this->_signRequestParams( $requestParams );		

		return $requestParams;
	}

	/**
	 * Optional function: only needed for recurring payments:
	 * Returns subscription request parameters for gateway depending on basket (without specifying payment type)
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket   paymentBasket object
	 * @return array                                 Returns array $requestParams
	*/
	protected function getSubscriptionRequstParams( $paymentBasket )
	{
		// mandatory parameters:
		$requestParams						=   $this->_getBasicRequstParams( $paymentBasket );

		// Quickpay needs unique order_id add 's-s-' for subscription signup
		// always use period and amount 3 for form price and period if amount 1 is zero

		// check for subscription or if single payment:
		if ( $paymentBasket->period3 ) {
			$requestParams['order_id']		= 	's-s-'.$this->_prepareOrderID($paymentBasket->id, true );
			$requestParams['type']			=	'subscription';
			$requestParams['description']	=	$paymentBasket->item_name;
		}

		// sign subscription payment params:
		$this->_signRequestParams( $requestParams );
		
		return $requestParams;
	}

	/**
	* The user got redirected back from the payment service provider with a success message: let's see how successfull it was
	*
	* @param  cbpaidPaymentBasket  $paymentBasket  New empty object. returning: includes the id of the payment basket of this callback (strictly verified, otherwise untouched)
	* @param  array                $postdata       _POST data for saving edited tab content as generated with getEditTab
	* @return string                               HTML to display if frontend, text to return to gateway if notification, FALSE if XML error (and not yet ErrorMSG generated), or NULL if nothing to display
	*/
	protected function handleReturn( $paymentBasket, $postdata )
	{

		if ( ( count( $postdata ) > 0 ) && isset( $postdata['order_id'] ) ) {
			// we prefer POST for sensitive data:
			$requestdata			        =	$postdata;
		} else {
			// but if customer needs GET, we will work with it too (removing CMS/CB/CBSubs specific routing params):
			$requestdata			        =   $this->_getGetParams();
		}

		// check if order_id exists and if not add basket id and shared_secret from PDT success url:
		if ( ! isset( $requestdata['order_id'] ) ) {
			$requestdata['order_id']		=	$this->_prepareOrderID( cbGetParam( $_GET, 'cbpbasket', null ), true );
			$requestdata['shared_secret']           =   cbGetParam( $_GET, 'cbpid', null );
		}
		return $this->_returnParamsHandler( $paymentBasket, $requestdata, 'R',null);
	}

	/**
	* The user cancelled his payment
	*
	* @param  cbpaidPaymentBasket  $paymentBasket  New empty object. returning: includes the id of the payment basket of this callback (strictly verified, otherwise untouched)
	* @param  array                $postdata       _POST data for saving edited tab content as generated with getEditTab
	* @return string                               HTML to display if frontend, text to return to gateway if notification, FALSE if registration cancelled and ErrorMSG generated, or NULL if nothing to display
	*/
	protected function handleCancel( $paymentBasket, $postdata )
	{
		// The user cancelled his payment (and registration):
		if ( $this->hashPdtBackCheck( $this->_getReqParam( 'pdtback' ) ) ) {
			$paymentBasketId				=	(int) $this->_getReqParam( 'basket' );
			$shared_secret					=	$this->_getReqParam( 'id' );

			$exists							=	$paymentBasket->load( (int) $paymentBasketId );

			if ( $exists && ( $shared_secret == $paymentBasket->shared_secret ) && ( $paymentBasket->payment_status != 'Completed' ) ) {

				//As Quickpay allready has stored the order_id, it is taken, so we need to delete the basket to start a new order
				$paymentBasket->delete($paymentBasketId);

				$this->_setErrorMSG( CBPTXT::T( 'Payment cancelled.' ) );

				//Need to reload page using Joomla core, to get a new basket and show Cancel message
				$app = JFactory::getApplication();
				$app->enqueueMessage(CBPTXT::T( "Payment cancelled." ), 'notice'); //Translate cancel message with CBSubs and deliver to Joomla alert message
				$app->redirect(JRoute::_('index.php?option=com_comprofiler&view=userprofile')); //Show userprofile again
			}
		}

		return 'Payment cancelled.';
	}

	/**
	* The payment service provider server did a server-to-server notification: verify and handle it here:
	*
	* @param  cbpaidPaymentBasket  $paymentBasket  New empty object. returning: includes the id of the payment basket of this callback (strictly verified, otherwise untouched)
	* @param  array                $postdata       _POST data for saving edited tab content as generated with getEditTab
	* @return string                               HTML to display if frontend, text to return to gateway if notification, FALSE if registration cancelled and ErrorMSG generated, or NULL if nothing to display
	*/
	protected function handleNotification( $paymentBasket, $postdata )
	{	
		$rawpostdata			            =	file_get_contents( 'php://input' );

		if ( $rawpostdata && ( ( $rawpostdata[0] === '{' ) || ( $rawpostdata[0] === '[' ) ) ) {
			$postdata			            =	json_decode( $rawpostdata, true, 512, JSON_UNESCAPED_UNICODE | JSON_UNESCAPED_SLASHES | JSON_NUMERIC_CHECK );
		}

		if ( ( count( $postdata ) > 0 ) && isset( $postdata['order_id'] ) ) {
			// we prefer POST for sensitive data:
			$requestdata		            =	$postdata;
		} else {
			// but if gateway needs GET, we will work with it too:
			$requestdata		            =	$this->_getGetParams();
		}
		//Check if notification is valid with API key
		$valid = ( $_SERVER["HTTP_QUICKPAY_CHECKSUM_SHA256"] == hash_hmac( 'sha256', file_get_contents("php://input"), $this->getAccountParam( 'psp_key' ) ) );

		return $this->_returnParamsHandler( $paymentBasket, $requestdata, 'I', $valid );
	}


	/**
	 * Attempts to authorize and capture a credit card for a single payment of a payment basket using PSP DirectLink
	 *
	 * @param  cbpaidPaymentBasket  $paymentBasket
	 * @param  string               $returnText                  RETURN param
	 * @param  boolean              $transientErrorDoReschedule  RETURN param
	 * @return boolean              TRUE: succes, FALSE: failed or unknown result
	 */
	protected function processScheduledAutoRecurringPayment( $paymentBasket, &$returnText, &$transientErrorDoReschedule )
	{
		global $_CB_framework;

		$error = null;
		//Call API to make recurring payment
		$fields_array 						= 	array(
				'order_id'                  => "s-p".( ( $paymentBasket->recur_times_used ) + 2 )."-".$this->_prepareOrderID( $paymentBasket->id, true ),
				'amount'                    => intval( sprintf( '%2.f', $paymentBasket->mc_amount3 ) * 100 ),
				'auto_capture'              => true
													);
		$error                              =   $this->httpsRequestGuzzle($this->_pspApiUrl() . '/subscriptions/' . $paymentBasket->subscr_id .'/recurring',$fields_array,105,$response_recurring,$status,'post','json','',true,'','',$this->getAccountParam( 'psp_api_user_key' ) ,false,null,$this->gatewayApiVersion);
		if (!$error && $status = '202')
		{
			$error                          = null;
			//Check recurring payment status through API
			$error                              =   $this->httpsRequestGuzzle($this->_pspApiUrl() . '/payments/'.$response_recurring['id'],$fields_array,105,$response,$status,'get','json','',true,'','',$this->getAccountParam( 'psp_api_user_key' ) ,false,null,$this->gatewayApiVersion);
		}
		if ( $error || ( $status != 200 ) || ! $response ) {
			// An error occurred or got returned: Log it as $ipn notification with error:
			$now					    	=	$_CB_framework->now();
			$ipn						    =	$this->_prepareIpn( 'B', $paymentBasket->payment_status, 'Autorecurring', 'PSP DataLink unavailable', $now, 'utf-8' );

			$ipn->bindBasket( $paymentBasket );

			$errorText						=	'HTTPS POST Connection to payment gateway server failed (check system information in CBSubs Settings): ERROR: ' . $error . ' (' . ($status == -100 ? 'Timeout' : $status ) . ')';

			$ipn->setRawResult( $errorText );
			//$ipn->setRawData( '$response="' . addslashes( $response ) . '"' . "\n" . '$status=' . addslashes( $status ) . ';' . "\n" . '$xml=' . $datalink_url . ';' . "\n" . '$_POST=' . var_export( $formvars, true ) . ';' . "\n" );
			$this->_storeIpnResult( $ipn, 'APILINKDOWN' );
			$this->_setLogErrorMSG( 4, $ipn, $this->getPayName() . ': ' . $errorText, CBPTXT::T( "Sorry, the payment server did not reply." ) . ' ' . CBPTXT::T( "Please contact site administrator to check payment status and error log." ) );

			$user							=	CBuser::getUserDataInstance( $paymentBasket->user_id );
			$username						=	( $user ? $user->username : '?' );
			$returnText						=	sprintf( CBPTXT::T( "FAILED Auto-recurring payment of %s for basket %s and Order_id %s of %s (username %s - user id %s) using %s due to error %s." ), $paymentBasket->renderPrice( $paymentBasket->mc_amount3, null, null, false ), $paymentBasket->id, $paymentBasket->sale_id, $paymentBasket->first_name . ' ' . $paymentBasket->last_name, $username, $paymentBasket->user_id, $this->getPayName(), ' - HTTP error ' . ': ' . $error . ' ' . 'Status' . ': ' . $status );
			$transientErrorDoReschedule		=	true;
			$return							=	false;
		} else {
			//Handle result
			$paymentResult					=   $this->_returnParamsHandler( $paymentBasket, $response, 'A', '' );
			$user							=	CBuser::getUserDataInstance( $paymentBasket->user_id );
			$username						=	( $user ? $user->username : '?' );

			if ( $paymentResult !== false ) {
				if ( ( $paymentResult === true ) && in_array( $paymentBasket->payment_status, array( 'Completed', 'Pending' ) ) ) {
					if ( $paymentBasket->payment_status == 'Completed') {
						$returnText				=	sprintf( CBPTXT::T( "Completed Auto-recurring payment of %s for basket %s and Order_id %s of %s (username %s - user id %s), using %s with txn_id %s and auth_id %s." ), $paymentBasket->renderPrice(  $paymentBasket->mc_amount3, null, null, false ), $paymentBasket->id, $paymentBasket->sale_id, $paymentBasket->first_name . ' ' . $paymentBasket->last_name, $username, $paymentBasket->user_id, $this->getPayName(), $paymentBasket->txn_id, $paymentBasket->auth_id );
					} else {
						$returnText				=	sprintf( CBPTXT::T( "Pending Auto-recurring payment of %s for basket %s and Order_id %s of %s (username %s - user id %s), using %s with txn_id %s and auth_id %s for reason: %s." ), $paymentBasket->renderPrice( $paymentBasket->mc_amount3, null, null, false ), $paymentBasket->id, $paymentBasket->sale_id, $paymentBasket->first_name . ' ' . $paymentBasket->last_name, $username, $paymentBasket->user_id, $this->getPayName(), $paymentBasket->txn_id, $paymentBasket->auth_id, $paymentBasket->reason_code );
					}

					$transientErrorDoReschedule	=	false;
					$return						=	true;
				} else {
					$returnText					=	sprintf( CBPTXT::T( "FAILED (%s) Auto-recurring payment of %s for basket %s and Order_id %s of %s (username %s - user id %s), using %s due to error %s." ), $paymentBasket->payment_status, $paymentBasket->renderPrice( $paymentBasket->mc_amount3, null, null, false ), $paymentBasket->id, $paymentBasket->sale_id, $paymentBasket->first_name . ' ' . $paymentBasket->last_name, $username, $paymentBasket->user_id, $this->getPayName(), $paymentBasket->reason_code );
					$transientErrorDoReschedule	=	true;
					$return						=	false;
				}
			} else {
				$returnText						=	sprintf( CBPTXT::T( "FAILED (Error) Auto-recurring payment of %s for basket %s and Order_id %s of %s (username %s - user id %s), using %s due to error %s." ), $paymentBasket->renderPrice( $paymentBasket->mc_amount3, null, null, false ), $paymentBasket->id, $paymentBasket->sale_id, $paymentBasket->first_name . ' ' . $paymentBasket->last_name, $username, $paymentBasket->user_id, $this->getPayName(), $paymentBasket->reason_code );
				$transientErrorDoReschedule		=	true;
				$return							=	false;
			}
		}
		return $return;
	}


	/**
	 * GATEWAY-INTERNAL SPECIFIC PRIVATE METHODS:
	 */

	/**
	 * gives gateway API URL server name from gateway URL list
	 *
	 * @return string  server-name (with 'https://' )
	 */
	private function _pspApiUrl( )
	{
		return $this->gatewayUrl( 'psp+api' );
	}


	/**
	 * Cancels a subscription
	 *
	 * @param  cbpaidPaymentBasket   $paymentBasket  paymentBasket object
	 * @param  cbpaidPaymentItem[]   $paymentItems   redirect immediately instead of returning HTML for output
	 * @return boolean                               true if unsubscription done successfully, false if error
	 */
	protected function handleStopPaymentSubscription( $paymentBasket, $paymentItems )
	{
		global $_CB_framework;
		$return									=	false;

		if ( $paymentBasket->mc_amount3 ) {

			// Recurring amount existing and if first amount existed it got payed OK:
			// Stop CBSubs scheduled auto-payments:
			// only if an actual subscription id exists:
			if ( $paymentBasket->subscr_id ) {

				$fields_array = array(
					'id'                       => $paymentBasket->subscr_id,
					'synchrinized'             => true
				);
				$error =	$this->httpsRequestGuzzle($this->_pspApiUrl() .'/subscriptions/'.$paymentBasket->subscr_id .'/cancel',$fields_array,105,$response,$status,'post','json','',true,'','',$this->getAccountParam( 'psp_api_user_key' ) ,false,null,$this->gatewayApiVersion);

				if ( $error || ( $status != 202 ) || ( ! $response ) ) {
					$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': unsubscribe API response failed', CBPTXT::T( 'Submitted unsubscription failed on-site.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );
				} else {
					if ( $response['accepted'] == true ) {
						// Stop CBSubs scheduled auto-payments: Is this enough !!!????
						//$paymentBasket->unscheduleAutoRecurringPayments();
						$ipn					=	$this->_prepareIpn( 'R', $paymentBasket->payment_status, $paymentBasket->payment_type, 'Unsubscribe', $_CB_framework->now(), 'utf-8' );
						$ipn->test_ipn			=	$paymentBasket->test_ipn;
						$ipn->raw_result		=	'SUCCESS';
						$ipn->raw_data			=	'$message_type="STOP_PAYMENT_SUBSCRIPTION"' . ";\n"
							.	/* cbGetParam() not needed: we want raw info */ '$response=' . var_export( $response, true ) . ";\n"
							.	/* cbGetParam() not needed: we want raw info */ '$_GET=' . var_export( $_GET, true ) . ";\n"
							.	/* cbGetParam() not needed: we want raw info */ '$_POST=' . var_export( $_POST, true ) . ";\n";

						$ipn->bindBasket( $paymentBasket );

						$ipn->sale_id			=	$paymentBasket->id;
						$ipn->txn_id			=	$paymentBasket->txn_id;
						$ipn->subscr_id			=	$paymentBasket->subscr_id;
						$ipn->txn_type			=	'subscr_cancel';

						$this->_storeIpnResult( $ipn, 'SUCCESS' );
						$this->_bindIpnToBasket( $ipn, $paymentBasket );

						$return					=	true;
					} else {
						$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': ' . $response, CBPTXT::T( 'Sorry, the payment server replied with an error.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check payment status and error log.' ) );
					}
				}
			} else {
				$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': unsubscribe failed from missing subscr_id in payment basket', CBPTXT::T( 'Submitted unsubscription failed on-site.' ) . ' ' . CBPTXT::T( 'Please contact site administrator to check error log.' ) );
			}
		}
		return $return;
	}

	/**
	 * GATEWAY-INTERNAL SPECIFIC PRIVATE METHODS:
	 */


	/**
	 * sign payment request $requestParams with validation code added to $requestParams array
	 *
	 * @param  array     $requestParams        The keyed parameters array to add generate validation to $requestParams['signature']
	 */
	private function _signRequestParams( &$requestParams )
	{
		// Concatenate them using this payments concatenation function with $caseInsensitiveKeys = true:
		$string						=	$this->_concatVars( $requestParams, array(), null, ' ', '', true, true, false, true );		// add validation code doing sha256 hash of the string:
		$requestParams['checksum']	=	hash_hmac( 'sha256', $string, $this->getAccountParam( 'psp_api_key' ) );
	}


	/**
	 * Compute the CBSubs payment_status based on gateway's reply in $postdata:
	 *
	 * @param  array   $postdata  raw POST data received from the payment gateway
	 * @param  string  $reason    OUT: reason_code
     * @return string             CBSubs payment status
	 */
	private function _paymentStatus( $postdata, &$reason )
	{
		$operations					=	cbGetParam( $postdata, 'operations', array() );
		if ( ! $operations ) {
			$reason					=	null;
			return 'Error';
		}
		$latest						=	( count( $operations ) - 1 );
		$status						=	cbGetParam( $operations[$latest], 'qp_status_code' );

		switch ( $status ) {
			case '20000':
				switch ( cbGetParam( $operations[$latest], 'type' ) ) {
					case 'authorize':
						$reason		=	'Transaction being processed';
						$status		=	'Pending';
						break;
					case 'capture':
						$reason		=	null;
						$status		=	'Completed';
						break;
					case 'refund':
						$reason		=	null;
						$status		=	'Refunded';
						break;
					case 'recurring':
						if(cbGetParam( $operations[$latest], 'qp_status_msg' )=='Approved')
						{
							$reason		=	null;
							$status		=	'Completed';
						}
						else
						{
							$reason		=	'Recurring payment not passed!';
							$status		=	'Error';
						}

						break;
					default: // Unknown
						$reason		=	null;
						$status		=	null;
						break;
				}
				break;
			case '30100':
				$reason				=	'3D Secure is required';
				$status				=	'Denied';
				break;
			case '40000':
				$reason				=	'Rejected By Acquirer';
				$status				=	'Denied';
				break;
			case '40001':
				$reason				=	'Request Data Error';
				$status				=	'Error';
				break;
			case '50000':
				$reason				=	'Gateway Error';
				$status				=	'Error';
				break;
			case '50300':
				$reason				=	'Communications Error (with Acquirer)';
				$status				=	'Error';
				break;
		}
				
		return $status;
	}

	/**
	 * Compute the CBSubs payment_type based on gateway's reply $postdata:
	 *
	 * @param  array   $postdata raw POST data received from the payment gateway
	 * @return string  Human-readable string
	 */
	private function _getPaymentType( $postdata )
	{
		$metaData		=	cbGetParam( $postdata, 'metadata', array() );

		if ( ! $metaData ) {
			return null;
		}

		$type			=	strtolower( cbGetParam( $metaData, 'brand' ) );

		switch ( strtolower( $type ) ) {
			case 'american-express':
			case 'american-express-dk':
				$type	=	'American Express Credit Card';
				break;
			case 'dankort':
				$type	=	'Dankort Credit Card';
				break;
			case 'danske-dk':
				$type	=	'Danske Net Bank';
				break;
			case 'diners':
			case 'diners-dk':
				$type	=	'Diners Credit Card';
				break;
			case 'edankort':
				$type	=	'eDankort Credit Card';
				break;
			case 'fbg1886':
				$type	=	'Forbrugsforeningen af 1886';
				break;
			case 'jcb':
			case '3d-jcb':
				$type	=	'JCB Credit Card';
				break;
			case '3d-maestro':
			case '3d-maestro-dk':
				$type	=	'Maestro Credit Card';
				break;
			case 'mastercard':
			case 'mastercard-dk':
			case '3d-mastercard':
			case '3d-mastercard-dk':
				$type	=	'Mastercard Credit Card';
				break;
			case 'mastercard-debet-dk':
			case '3d-mastercard-debet-dk':
				$type	=	'Mastercard Debit Card';
				break;
			case 'nordea-dk':
				$type	=	'Nordea Net Bank';
				break;
			case 'visa':
			case 'visa-dk':
			case '3d-visa':
			case '3d-visa-dk':
				$type	=	'Visa Credit Card';
				break;
			case 'visa-electron':
			case 'visa-electron-dk':
			case '3d-visa-electron':
			case '3d-visa-electron-dk':
				$type	=	'Visa Debit Card';
				break;
			case 'mobilepay':
				$type	=	'MobilePay';
				break;
			case 'paypal':
				$type	=	'PayPal';
				break;
			case 'creditcard':
			case '3d-creditcard':
				$type	=	'Credit Card';
                break;
			default:
				break;
		}

		return $type;
	}
	





	/**
	* Popoulates basic request parameters for gateway depending on basket (without specifying payment type)
	*
	* @param  cbpaidPaymentBasket  $paymentBasket  paymentBasket object
	* @return array                                $requestParams
	*/
	private function _getBasicRequstParams( $paymentBasket )
	{
		global $_CB_framework;

		// mandatory parameters:
		$requestParams										=	array();
		$requestParams['version']							=	$this->gatewayApiVersion;
		$requestParams['merchant_id']						=	$this->getAccountParam( 'psp_id' );
		$requestParams['agreement_id']						=	$this->getAccountParam( 'psp_api_id' );
		$requestParams['order_id']							=	$this->_prepareOrderID($paymentBasket->id,true);
		$requestParams['amount']							=	( sprintf( '%2.f', $paymentBasket->mc_gross ) * 100 );
		$requestParams['currency']							=	$paymentBasket->mc_currency;
		$requestParams['type']								=	'payment';
		$requestParams['language']							=	$this->getAccountParam( 'language' );
		$requestParams['autocapture']						=	1;	

		// optional parameters:
		$requestParams['continueurl']						=	$this->getSuccessUrl( $paymentBasket );
		$requestParams['cancelurl']							=	$this->getCancelUrl( $paymentBasket );
		$requestParams['callbackurl']						=	$this->getNotifyUrl( $paymentBasket );

		// courtesy fields (pre-filled but editable on credit card mask):
		$requestParams['description']						=	$paymentBasket->item_name;
		$requestParams['variables[user_id]']				=	$paymentBasket->user_id;
		$requestParams['variables[basket_id]']				=	$paymentBasket->id;
		
		// recommended anti-fraud fields:
		$requestParams['invoice_address[name]']				=	$paymentBasket->address_name;

		if ( $this->getAccountParam( 'givehiddenemail' ) ||  $this->getAccountParam( 'givehiddenaddress' )){
			$requestParams['invoice_address_selection']     = true;
		}

		if ( $this->getAccountParam( 'givehiddenemail' ) && ( strlen( $paymentBasket->payer_email ) <= 50 ) ) {
			$requestParams['invoice_address[email]']		=	$paymentBasket->payer_email;
			$requestParams['customer_email']				=	$paymentBasket->payer_email;
		}

		if ( $this->getAccountParam( 'givehiddenaddress' ) ) {
			cbimport( 'cb.tabs' ); // needed for cbIsoUtf_substr()

			$addressFields									=	array(	'invoice_address[street]'			=>	array( $paymentBasket->address_street, 30 ),
																		'invoice_address[zip_code]'			=>	array( $paymentBasket->address_zip, 10 ),
																		'invoice_address[city]'				=>	array( $paymentBasket->address_city, 30 ),
																		'invoice_address[country_code]'		=>	array( $this->countryToLetters( $paymentBasket->address_country, 3 ), 3 )
																	);

			if ( $paymentBasket->address_state != 'other' ) {
				$addressFields['invoice_address[region]']	=	array( substr( $paymentBasket->address_state, -2 ), 2 );
			}

			foreach ( $addressFields as $k => $value_maxlength ) {
				$adrField									=	cbIsoUtf_substr( $value_maxlength[0], 0, $value_maxlength[1] );

				if ( $adrField ) {
					$requestParams[$k]						=	$adrField;
				}
			}
		}

		if ( $this->getAccountParam( 'givehiddentelno' ) && ( strlen( $paymentBasket->contact_phone ) <= 50 ) ) {
			$requestParams['invoice_address[phone_number]']	=	$paymentBasket->contact_phone;
		}
		return $requestParams;
	}



    /**
     * The user got redirected back from the payment service provider with a success message: let's see how successfull it was
     *
     * @param  cbpaidPaymentBasket  $paymentBasket       New empty object. returning: includes the id of the payment basket of this callback (strictly verified, otherwise untouched)
     * @param  array                $requestdata         Data returned by gateway
     * @param  string               $type                Type of return ('I' for INS, 'R' for PDT, 'P' for API (first payment) 'A' for Autorecurring payment (API) )
     * @param  null                 $additionalLogData   Additional Texts to log into IPN raw data
     * @return string                                    HTML to display if frontend, text to return to gateway if notification, FALSE if registration cancelled and ErrorMSG generated, or NULL if nothing to display
     */
	private function _returnParamsHandler( $paymentBasket, $requestdata, $type, $additionalLogData = null, $validcall = null )
	{
		global $_CB_framework, $_GET, $_POST;
		$ret													=	null;
		$paymentBasketId										=	(int) $this->_prepareOrderID(cbGetParam( $requestdata, 'order_id',null),null);

		if ( $paymentBasketId ) {
			$exists												=	$paymentBasket->load( (int) $paymentBasketId );

			if ( isset( $requestdata['shared_secret'] ) ) {
				if ( $requestdata['shared_secret'] == $paymentBasket->shared_secret ) {
					$validSharedSecret							=	true;
				} else {
					$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': The shared_secret is missing in $_GET.', CBPTXT::T( "Please contact site administrator to check error log." ) );
					$validSharedSecret							=	false;
				}
			} else {
				$validSharedSecret								=	null;
			}

			// 1) Payment basket exists
			// 2) It's not a Returned-result to site with a missing or invalid shared secret in the cookie
			// 3) a) It's an automatic payment
			//    OR b) the returned 'id' matches the basket secret
			//		    but it's Not a Returned-result for an already completed basket:


			if ( $exists
			&& ( ! ( ( $type == 'R' ) && ( $validSharedSecret !== true ) ) )
				&&	( ( ( $type == 'P' ) || ( $type == 'A' ) )
					||  ( ! ( ( $type == 'R' ) && ( $paymentBasket->payment_status == 'Completed' ) ) ) ) )
			{
				if($type == 'I'){ // If callback type
					if ( cbGetParam( $requestdata, 'type',null ) == 'Subscription' ) //If is callback from subscription signup just ignore and return
					{
						return true;
					}
					//If is Callback other than subscription sign up, set paymentStatus
					$paymentStatus									=	$this->_paymentStatus( $requestdata, $reason );
				}

				if($type=='A') //Is recurring type
				{
					$requestdata['paymenttype'] = null;
					$paymentStatus									=	$this->_paymentStatus( $requestdata, $reason );
				}

				// Log the return record:
				$log_type										=	$type;
				$reason											=	null;

				if($type == 'R')
				{
					//Check requestdata - if this order_id is a singup order_id
					$this->httpsRequestGuzzle($this->_pspApiUrl() .'/subscriptions/?order_id='.'s-s-'.$this->_prepareOrderID(cbGetParam( $requestdata, 'order_id',null),true),null,105,$response,$status,'get','json','',true,'','',$this->getAccountParam( 'psp_api_user_key' ) ,false,null,$this->gatewayApiVersion);

					if(cbGetParam($response,'id',0) != 0)
					{
						//Find subscription_id from signup__orderid and make first recurring payment
						$order_id = 's-s-'.$this->_prepareOrderID(cbGetParam( $requestdata, 'order_id',null),true);
						//Get Subscription ID from Quickpay through API from signup order_id
						$error =	$this->httpsRequestGuzzle($this->_pspApiUrl() . '/subscriptions/?order_id='.$order_id, null,105,$response,$status,'get','json','',true,'','',$this->getAccountParam( 'psp_api_user_key' ) ,false,null,$this->gatewayApiVersion);
						$subscriptionID                         =   cbGetParam($response, 'id', 0);
						// Subscription ID found - make recurring payment
						if (!$error && $status = '200')
						{
							//remove "s-s-" from order_id
							$order_id                           =   str_replace("s-s-", "", $order_id);
							//Call API to make recurring payment
							$fields_array   = array(
								'order_id'                      =>  "s-p1-" . $order_id,
								'amount'                        =>  intval( sprintf( '%2.f', $paymentBasket->mc_gross ) * 100 ),
								'auto_capture'                  =>  true
							);
							$error =	$this->httpsRequestGuzzle($this->_pspApiUrl() . '/subscriptions/' . $subscriptionID . '/recurring', $fields_array,105,$response,$status,'post','json','',true,'','',$this->getAccountParam( 'psp_api_user_key' ) ,false,null,$this->gatewayApiVersion);
							if (!$error && $status = '202')
							{
								//Check recurring payment status through API
								$error =	$this->httpsRequestGuzzle($this->_pspApiUrl() . '/payments/?order_id=' . "s-p1-" . $order_id,null,105,$response,$status,'get','json','',true,'','',$this->getAccountParam( 'psp_api_user_key' ) ,false,null,$this->gatewayApiVersion);
							}
						}
						//Return reponse, error, status and set subscription_id to the id of the subscription signup order - to tie it together. The originally id of the first recurring payment is skipped!
						$response['order_id']                   =   $order_id;
						$response['subscription_id']            =   $subscriptionID;


						$requestdata                                    =   $response;
						$paymentStatus									=	$this->_paymentStatus( $requestdata, $reason );
						//The API sendt back an error upon signup
						if($error)
						{
							$paymentStatus = 'Error';
						}
						$requestdata['paymenttype'] 				= 	'subscription';
					}
					else
					//Is continueurl from singlepayment
					{
						//Get the status of the single payment from API call
						$error =	$this->httpsRequestGuzzle($this->_pspApiUrl() . '/payments/?order_id=' . $this->_prepareOrderID($requestdata['order_id'], true),null,105,$response,$status,'get','json','',true,'','',$this->getAccountParam( 'psp_api_user_key' ) ,false,null,$this->gatewayApiVersion);
						$requestdata = $response;
						$paymentStatus    = $this->_paymentStatus($requestdata, $reason);
						//The API sendt back an error upon making payment
						if ($error)
						{
							$paymentStatus = 'Error';
						}
						$requestdata['paymenttype']     = 'payment';
					}
				}

				$paymentType									=	$this->_getPaymentType( $requestdata );
				$paymentTime									=	$_CB_framework->now();

				if ( $paymentStatus == 'Error' ) {
					$errorTypes									=	array( 'I' => 'D', 'R' => 'E', 'P' => 'V', 'A' => 'W' );

					if ( isset( $errorTypes[$type] ) ) {
						// Converts error status to CBSubs's error statuses:
						$log_type								=	$errorTypes[$type];
					}
				}

				$ipn											=	$this->_prepareIpn( $log_type, $paymentStatus, $paymentType, $reason, $paymentTime, 'utf-8' );

				if ( $paymentStatus == 'Refunded' ) {
					// in case of refund we need to log the payment as it has same TnxId as first payment: so we need payment_date for discrimination:
					$ipn->payment_date							=	gmdate( 'H:i:s M d, Y T', $paymentTime ); // paypal-style
				}

				$ipn->test_ipn									=	0;
				$ipn->raw_data									=	'$message_type="' . ( $type == 'R' ? 'RETURN_TO_SITE' : ( $type == 'I' ? 'NOTIFICATION' : ( $type == 'P' ? 'DATALINK PAYMENT' : ( $type == 'A' ? 'AUTORECURRING DIRECTLINK PAYMENT' : 'UNKNOWN' ) ) ) ) . '";' . "\n";

				if ( $additionalLogData ) {
					foreach ( $additionalLogData as $k => $v ) {
						$ipn->raw_data							.=	'$' . $k . '="' . var_export( $v, true ) . '";' . "\n";
					}
				}

				$ipn->raw_data									.=	/* cbGetParam() not needed: we want raw info */ '$requestdata=' . var_export( $requestdata, true ) . ";\n"
																.	/* cbGetParam() not needed: we want raw info */ '$_GET=' . var_export( $_GET, true ) . ";\n"
																.	/* cbGetParam() not needed: we want raw info */ '$_POST=' . var_export( $_POST, true ) . ";\n";
				
				if ( $paymentStatus == 'Error' ) {
					$paymentBasket->reason_code					=	$reason;

					$this->_storeIpnResult( $ipn, 'ERROR:' . $reason );
					$this->_setLogErrorMSG( 4, $ipn, $this->getPayName() . ': ' . $reason, CBPTXT::T( "Sorry, the payment server replied with an error." ) . ' ' . CBPTXT::T( "Please contact site administrator to check payment status and error log." ) );

					$ret										=	false;
				} else {		
					// Payment status is not Error:
					// Get important parameters from basket into the $ipn:
					$ipn->bindBasket( $paymentBasket );
					
					//Prepare orderID
					$requestdata['order_id'] = $this->_prepareOrderID( cbGetParam( $requestdata,'order_id',null),true);
					//Set the txn_id to the id of the order (Not the same as order_id !!
					$requestdata['txn_id'] = cbGetParam( $requestdata, 'id',null );
					
					// Convert fields from the gateway INS to CB's $ipn Notification format:
					$insToIpn									=	array(
																			'mc_currency' => 'currency',
																			'sale_id' => 'order_id',
																			'txn_id' => 'txn_id',
																			'first_name' => 'customer_fname',
																			'last_name' => 'customer_lname',
																			'address_street' => 'address1',
																			'address_zip' => 'zipcode',
																			'address_city' => 'city',
																			'address_country' => 'country',
																			'address_state' => 'state',
																			'payer_email' => 'email',
																			'contact_phone' => 'phone_number',
																			'ip_addresses' => 'ip_address'
																		);

					foreach ( $insToIpn as $k => $v ) {
						$ipn->$k								=	cbGetParam( $requestdata, $v );
					}
					
					$operation 									= 	cbGetParam( $requestdata, 'operations', array());

					if(!is_array($operation[0]))
					{
						$ipn->mc_gross							= 	sprintf( '%2.2f',(cbGetParam( $requestdata, 'amount',0 ) / 100 ) );
					}else{
						$ipn->mc_gross								=	sprintf( '%2.2f', ( cbGetParam( $operation[0], 'amount',0 ) / 100 ) );
					}
					
					$ipn->user_id								=	(int) $paymentBasket->user_id;
	
					$recurring									=	( $type == 'A' || (($requestdata['paymenttype']=='subscription' && $type == 'R')) ? true : false );

					if ($recurring) {
						if ( $type != 'A' ) {
							// in case of first payment of auto-recurring payments:
							$ipn->txn_type						=	'subscr_signup';
							$ipn->subscr_id						=	cbGetParam( $requestdata, 'subscription_id',0 );
							$ipn->subscr_date					=	$ipn->payment_date;
							if($paymentBasket->mc_amount1 != $paymentBasket->mc_amount3){ //First payment is different from the next
								$ipn->period2    = $ipn->period1 =  $paymentBasket->period1;
								$ipn->mc_amount2 = $ipn->mc_amount1 =$paymentBasket->mc_amount1;
								$ipn->amount2    = $ipn->amount1= $paymentBasket->mc_amount1;
							}
							else
							{	//First payment is same as the next
								$ipn->period3    = $paymentBasket->period3;
								$ipn->mc_amount3 = $paymentBasket->mc_amount3;
								$ipn->amount3    = $paymentBasket->mc_amount3;
							}
							$ipn->recurring = $paymentBasket->recurring;
							$ipn->reattempt = $paymentBasket->reattempt;

						} else {
							if ( $paymentStatus == 'Denied' ) {
								// Auto-recurring payment got denied:
								if ( ( $paymentBasket->reattempts_tried + 1 ) <= cbpaidScheduler::getInstance( $this )->retries ) {
									// Schedule a new trial is possible: log it as simply a failed attempt:
									$ipn->txn_type				=	'subscr_failed';
								} else {
									// Schedule a new trial is not possible anymore: log it as a cancellation (?):
									$ipn->txn_type				=	'subscr_cancel';
								}
							} elseif ( in_array( $paymentStatus, array( 'Completed', 'Processed', 'Pending' ) ) ) {
								// Normal case: it's a recurring payment:
								$ipn->txn_type					=	'subscr_payment';
								//Test !!!!
								$ipn->period1 = $paymentBasket->period1;
								$ipn->period3 = $paymentBasket->period3;
								$ipn->mc_amount3 = $paymentBasket->mc_amount3;
								$ipn->amount3 = $paymentBasket->mc_amount3;
								$ipn->recurring = $paymentBasket->recurring;
								$ipn->reattempt = $paymentBasket->reattempt;
							}
						}
					} else {
						// Single payment:
						$ipn->txn_type							=	'web_accept';
									}
					

					// API Payments and Auto-recurring payments do not have a hash signature, other ones do have one that must be checked:
					if ( in_array( $type, array( 'P', 'A') ) || $validcall  || ($type == 'R' )) {
						// 1) Basket id matches orderId
						// 2) a) Amounts match
						//	  b) OR it's @author brunner
						// 3) Currency matches:

						if	( ( $paymentBasketId == (int) $this->_prepareOrderID( cbGetParam( $requestdata,'order_id',null), null )) 
							&& (	( sprintf( '%.2f', $paymentBasket->mc_gross ) == $ipn->mc_gross ) || ( sprintf( '%.2f', $paymentBasket->mc_amount3 ) == $ipn->mc_gross )
								||	( $ipn->payment_status == 'Refunded' ) )
							&& ( $paymentBasket->mc_currency == $ipn->mc_currency ) )
						{
							// Ok, we passed all basic consistancy tests:
							if ( in_array( $ipn->payment_status, array( 'Completed', 'Processed', 'Pending', 'Refunded', 'Denied' ) ) ) {
								// Status is also that of a successful (or on it's way to successful) transaction:
								// Stores the $ipn notification as a successful one:
								$this->_storeIpnResult( $ipn, 'SUCCESS' );
								// Binds to basket the relevant $ipn attributes:
								$this->_bindIpnToBasket( $ipn, $paymentBasket );

								// Sets autorecur and autorenew types:
								// 0: not auto-recurring, 1: auto-recurring without payment processor notifications, 2: auto-renewing with processor notifications updating $expiry_date:
								$autorecurring_type				=	( in_array( $ipn->txn_type, array( 'subscr_payment', 'subscr_signup', 'subscr_modify', 'subscr_eot', 'subscr_cancel', 'subscr_failed' ) ) ? 2 : 0 );

								// 0: not auto-renewing (manual renewals), 1: asked for by user, 2: mandatory by configuration:
								$autorenew_type					=	( $autorecurring_type ? ( ( $this->getAccountParam( 'enabled', 0 ) == 3 ) && ( $paymentBasket->isAnyAutoRecurring() == 2 ) ? 1 : 2 ) : 0 );

								if ( $recurring ) {
									// We want to reattempt auto-recurring payment in case of failure:
									$paymentBasket->reattempt	=	1;
								}

								// In case of refund we need to log the payment as it has same TnxId as first payment:
								$txnIdMultiplePaymentDates		=	( $paymentStatus == 'Refunded' );

								// Update payment basket, generates all basket payment events, generates a transaction record (payment):
								$this->updatePaymentStatus( $paymentBasket, $ipn->txn_type, $ipn->payment_status, $ipn, 1, $autorecurring_type, $autorenew_type, $txnIdMultiplePaymentDates );

								if ( in_array( $ipn->payment_status, array( 'Completed', 'Processed', 'Pending' ) ) ) {
									// ok, we had success in terms of a payment initiation or completion:
									$ret						=	true;

									if ( $recurring && ( $type != 'A' ) ) {

										// Now we need to schedule the automatic payments to be done by CBSubs using DirectLink, and which will trigger the auto-renewals (if it's not already the autorecurring payment (A)):
										$paymentBasket->scheduleAutoRecurringPayments();
									}
								}
							} else {
								// The payment didn't initiate or complete: We failed:
								// Store the notification $ipn as a failed one:
								$this->_storeIpnResult( $ipn, 'FAILED' );

								// Updates payment basket status with the notification status:
								$paymentBasket->payment_status	=	$ipn->payment_status;

								// Sets error to display to the user:
								$this->_setErrorMSG( '<div class="alert alert-info">' . $this->getTxtNextStep( $paymentBasket ) . '</div>' );

								// We will be re-displaying the basket to the user so he can retry payment:
								$paymentBasket->payment_status	=	'RedisplayOriginalBasket';
								$ret							=	false;
							}
						} else {
							// We found a major mismatch between orderID, amount or currency:
							// Store it in notification:
							$this->_storeIpnResult( $ipn, 'MISMATCH' );
							// Log it as error and prepare display of error to user:
							$this->_setLogErrorMSG( 3, $ipn, $this->getPayName() . ': orderID, amount or currency missmatch.', CBPTXT::T( "Sorry, the payment does not match the basket." ) . ' ' . CBPTXT::T( "Please contact site administrator to check error log." ) );

							$ret								=	false;
						}
					} else {
						// MD5 hash mismatches:
						// Store it in notification:
						$this->_storeIpnResult( $ipn, 'SIGNERROR' );
						// Log it as error and prepare display of error to user:
						$this->_setLogErrorMSG( 3, $ipn, $this->getPayName() . ': Quickpay e-Commerce MD5 Salt hash does not match with gateway. Please check MD5 Salt setting.', CBPTXT::T( "The MD5 signature is incorrect." ) . ' ' . CBPTXT::T( "Please contact site administrator to check error log." ) );

						$ret									=	false;
					}
				}
			}
		} else {
			// No basket found: 
			// Log it as error and prepare display of error to user:
			$this->_setLogErrorMSG( 3, null, $this->getPayName() . ': The OrderId is missing in the return URL: Probably due to missing "cbsubs_quickpay_basket" cookie', CBPTXT::T( "Please contact site administrator to check error log." ) );
		}

		return  $ret;
	}	// end function _returnParamsHandler


	/**
	 * Output properly formatted basket id number from order_id or generate order_id from raw basket id number
	 *
	 * @param  int      $orderId (Is build from this pattern :	| Normal payment: -{your prefix}- | Subscription signup: s-s--{your prefix}- | Recurring payment: s-p{count}--{your prefix}- )
	 * @param  boolean  $append (If $append=true, get raw order number and return like this -{your prefix}-{order number}, if $append=flase just return raw order number)
	 * @return string
	 */
	private function _prepareOrderID( $orderId, $prepend = false )
	{
		if ( $prepend ) { //First remove prepended text and order prefix to be shure it is raw order number then prepend prefix
			if ((substr($orderId,0,3) == 's-s') || (substr($orderId,0,3) == 's-p'))
			{ //Is recurring payment or sign up
				$returnstring = str_replace("-" . $this->_sanitizedPrefix() . "-", '', $orderId);
				$returnarray  = explode("-", $returnstring);
				//Now prepend order prefix
				return "-". $this->_sanitizedPrefix() . "-" . $returnarray[2]; //sanitize prefix and limit to 8 characters
			}
			else{ //If is not sign up og recurring, be sure not to prepend twice
				$returnstring = str_replace("-".$this->getAccountParam('psp_prefix')."-", '', $orderId);
				 return "-" . $this->_sanitizedPrefix() . "-" . $returnstring; //Sanitize prefix and limit to 8 characters
			}
		} else { //Get raw order number
			if ((substr($orderId,0,3) == 's-s') || (substr($orderId,0,3) == 's-p'))
			{ //Is recurring payment or sign up
				$returnstring = str_replace("-" . $this->_sanitizedPrefix() . "-", '', $orderId);
				$returnarray  = explode("-", $returnstring);
				return $returnarray[2];
			}
			else{ //If is not sign up og recurring, be sure not to prepend twice
				$returnstring = str_replace("-" . $this->_sanitizedPrefix() . "-", '', $orderId);
				return  $returnstring;
			}
		}
	}

	/**
	 * Output sanitized order prefix: only alpha and - and maximum 8 characters long
	 *
	 * @return string
	 */

	private function _sanitizedPrefix(){
		return substr(preg_replace('/[^-a-zA-Z0-9]/','',$this->getAccountParam('psp_prefix')),0,8);
	}


	/**
	 * USED by XML interface ONLY !!! Renders URL to set in the QuickPay interface for notifications:
	 * Called by cbpaidGatewayAccountquickpayoem::renderNotifyUrl() just below in next class
	 *
	 * @param  string  $gatewayId
	 * @return string  HTML to display
	 */
	public function renderNotifyUrl( /** @noinspection PhpUnusedParameterInspection */ $gatewayId )
	{
		//FIXME : in such cases, the notification URL should not contain the gateway ID and be independant of the gateway ID but only of the gateway TYPE: That way multiple gateways for same PSP account can be setup (e.g. for conditions on promos)
		return $this->getNotifyUrl( null );
	}


	/**
	 * Posts a POST form by https or http and gets result.
	 *
	 * @param  string  $urlNoHttpsPrefix  URL without https:// in front (but works also with http:// or https:// in front, but it's ignored.
	 * @param  array|string  $formvars          Variables in form to post
	 * @param  int     $timeout           Timeout of the access
	 * @param  string  $result            RETURNING: the fetched access
	 * @param  int     $status            RETURNING: the status of the access (e.g. 200 is normal)
	 * @param  string  $getPostType       'post' (default) or 'get'
	 * @param  string  $contentType       'normal' (default) or 'xml' ($formvars['xml']=xml content) or 'json' (application/json)
	 * @param  string  $acceptType        '* / *' (default) or 'application/xml' or 'application/json'
	 * @param  boolean $https             SSL protocol (default: true)
	 * @param  int     $port              port number (default: 443)
	 * @param  string  $username          HTTP username authentication
	 * @param  string  $password          HTTP password authentication
	 * @param  boolean $allowHttpFallback Allow fallback to http if https not available locally (default: false)
	 * @param  string  $referer           referrer
	 * @param  string  $accept_version    Accept Version of API at PSP - mandatory for QuickPay
	 * @return int     $error             error-code of access (0 for ok)
	 */
	protected function httpsRequestGuzzle( $urlNoHttpsPrefix, $formvars, $timeout, &$result, &$status, $getPostType = 'post', $contentType='normal', $acceptType='*/*', $https = true, $port = 443, $username = '', $password = '', $allowHttpFallback = false, $referer = null, $accept_version = null ) {
		$urlNoHttpsPrefix					=	preg_replace( '/^https?:\/\//', '', $urlNoHttpsPrefix );

		try {
			if ( $port && ( ( $https && ( $port != 443 ) ) || ( ( ! $https ) && ( $port != 80 ) ) ) ) {
				$portPostfix				=	':' . (int) $port;
			} else {
				$portPostfix				=	'';
			}

			$url							=	( $https && ( ! $allowHttpFallback ) ? 'https://' : 'http://' ) . $urlNoHttpsPrefix . $portPostfix;

			$request						=	new \GuzzleHttp\Client();
			$headers						=	array();

			if ( $timeout ) {
				$headers['timeout']			=	(int) $timeout;
			}

			if ( $contentType == 'xml' ) {
				$headers['Content-Type']	=	'application/xml';
			} elseif ( $contentType == 'json' ) {
				$headers['Content-Type']	=	'application/json';
			} else {
				$headers['Content-Type']	=	'application/x-www-form-urlencoded';
			}

			if ( $username || $password ) {
				$headers['Authorization']	=	'Basic ' . base64_encode( $username . ':' . $password );
			}

			if ( $acceptType != '*/*' ) {
				$headers['Accept']			=	$acceptType;
			}

			if ( $referer ) {
				$headers['Referer']			=	$referer;
			}

			if ( $referer ) {
				$headers['Port']			=	$port;
			}

			if($accept_version) {
				$headers['Accept-Version']  = $accept_version;
			}

			$options						=	array();

			if ( $headers ) {
				$options['headers']			=	$headers;
			}

			switch ( strtoupper( $getPostType ) ) {
				case 'GET':
					if ( $formvars ) {
						$options['query']	=	$formvars;
					}

					$results				=	$request->get( $url, $options );
					break;
				case 'POST':
				default:
					if ( $formvars ) {
						if ( version_compare( \GuzzleHttp\ClientInterface::VERSION, '6.0.0', '>=' ) ) {
							$options['form_params']		=	$formvars;
						} else {
							$options['body']			=	$formvars;
						}
					}

					$results				=	$request->post( $url, $options );
					break;
			}

			$status							=	$results->getStatusCode();

			if ( !($status == 200 || $status == 202 )) {
				$error						=	$results->getReasonPhrase();
			} else {
				$error						=	null;
			}
			$result_array					=	json_decode((string) $results->getBody(),true);
			if(is_array($result_array)){
				if(sizeof($result_array) == 1){
					$result                         =   $result_array[0];
				}else
				{
					$result                         =   $result_array;
				}
			}
			else{
				$result = $results->getReasonPhrase();
			}

		} catch( Exception $e ) {
			$error							=	$e->getMessage();
		}

		return $error;
	}
}

/**
 * Payment account class for this gateway: Stores the settings for that gateway instance, and is used when editing and storing gateway parameters in the backend.
 *
 * OEM base
 * No methods need to be implemented or overriden in this class, except to implement the private-type params used specifically for this gateway:
 */
class cbpaidGatewayAccountquickpayoem extends cbpaidGatewayAccounthostedpage		//  extends cbpaidGatewayAccount
{
	/**
	 * USED by XML interface ONLY !!! Renders URL for notifications
	 *
	 * @param  string           $gatewayId  Id of gateway
	 * @param  ParamsInterface  $params     Params of gateway
	 * @return string                       HTML to display
	 */
	public function renderNotifyUrl( $gatewayId, /** @noinspection PhpUnusedParameterInspection */ $params )
	{
		$payClass	=	$this->getPayMean();

		return str_replace( 'http://', 'https://', $payClass->renderNotifyUrl( $gatewayId ) );
	}

	/**
	 * USED by XML interface ONLY !!! Renders URL for site returns
	 *
     * @param  string           $gatewayId  Id of gateway
     * @param  ParamsInterface  $params     Params of gateway
     * @return string                       HTML to display
     */
	public function renderSiteUrl( $gatewayId, $params )
	{
		global $_CB_framework;

		return $_CB_framework->getCfg( 'live_site' );
	}

	/**
	* Gets payment mean handler : Overridde to phpdocument return of correct class
	*
	* @param  string             $methodCheck
	* @return cbpaidquickpayoem
	*/
	public function getPayMean( $methodCheck = null )
	{
		return parent::getPayMean( $methodCheck );
	}
}



/**
 * Payment handler class for this gateway: Handles all payment events and notifications, called by the parent class:
 *
 * This is the gateway-specific version.
 * Please note that except the constructor and the API version this class does not implement any public methods.
 */
class cbpaidquickpay extends cbpaidquickpayoem
{
}

/**
 * Payment account class for this gateway: Stores the settings for that gateway instance, and is used when editing and storing gateway parameters in the backend.
 *
 * This is the gateway-specific version.
 * No methods need to be implemented or overriden in this class, except to implement the private-type params used specifically for this gateway:
 */
class cbpaidGatewayAccountquickpay extends cbpaidGatewayAccountquickpayoem
{
}
